//mixerstore.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"

#define _MAX_MIXERS 8

struct savannahice_mixerstore {
 size_t refc;
 size_t mixers_len;
 savannahice_mixer_t mixers[_MAX_MIXERS];
};

savannahice_mixerstore_t savannahice_mixerstore_new(void) {
 savannahice_mixerstore_t ret = roar_mm_malloc(sizeof(struct savannahice_mixerstore));

 if ( ret == NULL )
  return NULL;

 memset(ret, 0, sizeof(struct savannahice_mixerstore));

 ret->refc = 1;
 ret->mixers_len = _MAX_MIXERS;

 return ret;
}

int savannahice_mixerstore_ref(savannahice_mixerstore_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc++;

 return 0;
}

int savannahice_mixerstore_unref(savannahice_mixerstore_t self) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc--;

 if (self->refc)
  return 0;

 for (i = 0; i < self->mixers_len; i++)
  if ( self->mixers[i] != NULL )
   savannahice_mixer_unref(self->mixers[i]);

 roar_mm_free(self);

 return 0;
}

int savannahice_mixerstore_add_mixer(savannahice_mixerstore_t self, savannahice_mixer_t mixer) {
 size_t i;

 if ( self == NULL || mixer == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < self->mixers_len; i++) {
  ROAR_DBG("savannahice_mixerstore_add_mixer(self=%p, mixer=%p): i=%zu", self, mixer, i);
  if ( self->mixers[i] == NULL ) {
   if ( savannahice_mixer_ref(mixer) == -1 )
    return -1;
   self->mixers[i] = mixer;
   return 0;
  }
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

savannahice_mixer_t savannahice_mixerstore_get_mixer_by_id(savannahice_mixerstore_t self, int id) {
 savannahice_stream_t stream;
 int streamid;
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 for (i = 0; i < self->mixers_len; i++) {
  if ( self->mixers[i] != NULL ) {
   stream = savannahice_mixer_stream(self->mixers[i]);
   streamid = savannahice_stream_get_id(stream);
   savannahice_stream_unref(stream);
   if ( id == -1 || id == streamid ) {
    if ( savannahice_mixer_ref(self->mixers[i]) == -1 )
     return NULL;
    return self->mixers[i];
   }
  }
 }

 roar_err_set(ROAR_ERROR_NOENT);
 return NULL;
}

int savannahice_mixerstore_handle(savannahice_mixerstore_t self) {
 int_least32_t cur;
 size_t i;

 ROAR_DBG("savannahice_mixerstore_handle(self=%p) = ?", self);

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 ROAR_DBG("savannahice_mixerstore_handle(self=%p) = ?", self);

 for (i = 0; i < self->mixers_len; i++) {
  if ( self->mixers[i] != NULL ) {
   cur = savannahice_mixer_delay(self->mixers[i]);
   if ( cur < 2000 ) /* 2ms is just a random number. Feel free to find a better. */
    savannahice_mixer_handle(self->mixers[i]);
  }
 }

 ROAR_DBG("savannahice_mixerstore_handle(self=%p) = 0", self);
 return 0;
}

int_least32_t savannahice_mixerstore_delay(savannahice_mixerstore_t self) {
 int_least32_t min = -1;
 int_least32_t cur;
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < self->mixers_len; i++) {
  if ( self->mixers[i] != NULL ) {
   cur = savannahice_mixer_delay(self->mixers[i]);
   if ( min < 0 || cur < min ) {
    min = cur;
   }
  }
 }

 if ( min < 0 )
  roar_err_set(ROAR_ERROR_NOENT);

 return min;
}

//ll
