//savannahice.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"
#include <shout/shout.h>

enum pwmode {
 PWMODE_NONE = 0,
 PWMODE_ARG,
 PWMODE_DSTR,
 PWMODE_ASK
};

static savannahice_plugin_t  g_plugins = NULL;
savannahice_network_t g_network = NULL;
int g_verbose = ROAR_DBG_INFO_NONE;

savannahice_mixerstore_t g_mixerstore = NULL;

void usage (const char * progname) {
 fprintf(stderr, "Usage: %s [OPTIONS...]\n", progname);

 fprintf(stderr, "Misc Options:\n\n");
 fprintf(stderr,
        " --help                - Print this help.\n"
        " --verbose             - Be more verbose, can be used multiple times.\n"
       );

 fprintf(stderr, "\nPlugin Options:\n\n");
 fprintf(stderr,
        " --plugin-load FILE    - Load plugin FILE\n"
        " --plugin-args ARGS    - Arguments for the plugin\n"
        "                         (must be given before the --plugin-load)\n"
       );

 fprintf(stderr, "\nOutput Options:\n\n");
 fprintf(stderr,
        " -o  --odriver DRV     - Set the driver\n"
        " -O  --odevice DEV     - Set the device\n"
        " -oO OPTS              - Set output options\n"
        " -oN                   - Adds another output\n"
/* -oP                   - Mark output as primary */
        " --list-driver         - List all drivers (default driver: (autodetect))\n"
       );

 fprintf(stderr, "\nServer Options:\n\n");
 fprintf(stderr,
        " -u  --unix            - Use UNIX Socket\n"
        " -t  --tcp             - Use TCP Socket\n"
        " -n  --decnet          - Use DECnet Socket\n"
        "     --port PORT       - TCP port to bind to\n"
        "     --bind ADDR       - Nodename or path to bind to\n"
        "     --proto PROTO     - Use PROTO as protocol on Socket\n"
        "     --new-sock        - Parameters for new socket follow\n"
        "     --no-listen       - Do not listen for new clients\n"
        "                         (only useful for relaing)\n"
        "     --client-fh FH    - Comunicate with a client over this handle\n"
        "                         (only useful for relaing)\n"
        "     --close-fh FH     - Closes the given fh\n"
       );
}

static void print_driverlist(void) {
 printf("  Driver       Mode Flag Subsys - Description (devices)\n");
 printf("---------------------------------------------------------------\n");
 printf("  shout         w      V W      - libshout streaming (devices: http://user:pw@host:port/mount.ogg)\n");
}

void savannahice_exit(int val) {
 savannahice_plugin_unref(g_plugins);
 savannahice_network_unref(g_network);
 savannahice_mixerstore_unref(g_mixerstore);
 savannahice_stream_free();
 shout_shutdown();
 exit(val);
}

static const char * read_pw(struct roar_vio_calls * vio) {
 static char buffer[80];
 ssize_t len;

 len = roar_vio_read(vio, buffer, sizeof(buffer) - 1);

 if ( len == -1 )
  return NULL;

 // strip newlions.
 for (; buffer[len-1] == '\r' || buffer[len-1] == '\n'; len--);

 buffer[len] = 0;

 return buffer;
}

static int savannahice_add_mixer(char * opts, const char * device) {
 struct roar_vio_calls * vio;
 char * pwbuf = NULL;
 int ret;
 savannahice_mixer_t mixer = NULL;
 const char * server = NULL;
 uint16_t port = 0;
 const char * mount = NULL;
 const char * user = NULL;
 const char * password = NULL;
 enum pwmode pwmode = PWMODE_ARG;
 struct roar_keyval * kv = NULL, *c;
 ssize_t kvlen, i;
 char * a;
 const char * codec = NULL;

 // parser code stolen from roard/driver_shout.c.
 // This also ensures compatibility.

 if ( device != NULL ) {
  // device sould be an URL in this form:
  // [http[s]://][user[:pw]@]host[:port][/mp.ogg]

  if ( (a = strstr(device, "://")) != NULL ) {
   *a = 0;
   if ( strcmp(device, "http") ) {
    return -1;
   }
   device = a + 3;
  }

  // [user[:pw]@]host[:port][/mp.ogg]

  if ( (a = strstr(device, "@")) != NULL ) {
   *a = 0;
   user = device;
   device = a + 1;
  }

  if ( user != NULL ) {
   if ( (a = strstr(user, ":")) != NULL ) {
    *a = 0;
    password = a+1;
   }
  }

  if ( user != NULL && ! *user )
   user = NULL;

  if ( password != NULL && ! *password )
   password = NULL;

  // host[:port][/mp.ogg]

  if ( (a = strstr(device, "/")) != NULL ) {
   *a = 0;
   server = device;
   device = a + 1;
  } else {
   server  = device;
   device   += strlen(device);
  }

  if ( (a = strstr(server, ":")) != NULL ) {
   *a = 0;
   port = atoi(a+1);
  }

  if ( ! *server )
   server = NULL;

  // [/mp.ogg]

  if ( *device ) {
   mount = device;
  }
 }

 if ( opts != NULL ) {
  kvlen = roar_keyval_split(&kv, opts, ",", "=", 1);
  if ( kvlen == -1 )
   return -1;

/* options to support
 X -> codec
 X -> pwmode
 X -> user
 X -> password
 X -> pwfile
 * -> name
 * -> primary
 * Options to ignore:
 X -> rate, channels, bits
 X -> q
 X -> blocks, blocksize, mmap
 */

  for (i = 0; i < kvlen; i++) {
   c = &(kv[i]);
   if ( c->key == NULL )
    continue;

   // ignore some parameters to be compatible with roard:
   if ( !strcmp(c->key, "rate") || !strcmp(c->key, "channels") || !strcmp(c->key, "bits") ||
        !strcmp(c->key, "q") ||
        !strcmp(c->key, "blocks") || !strcmp(c->key, "blocksize") || !strcmp(c->key, "mmap") )
    continue;

   if ( !strcmp(c->key, "password") ) {
    password = c->value;
   } else if ( !strcmp(c->key, "user") ) {
    user = c->value;
   } else if ( !strcmp(c->key, "pwmode") ) {
    if ( !strcmp(c->value, "none") ) {
     pwmode = PWMODE_NONE;
    } else if ( !strcmp(c->value, "arg") || !strcmp(c->value, "para") || !strcmp(c->value, "parameter") ) {
     pwmode = PWMODE_ARG;
    } else if ( !strcmp(c->value, "ask") ) {
     pwmode = PWMODE_ASK;
    } else if ( !strcmp(c->value, "dstr") ) {
     pwmode = PWMODE_DSTR;
    } else {
     ROAR_WARN("Unknown password mode \"%s\", falling back to password-as-arg", c->value);
     pwmode = PWMODE_ARG;
    }
   } else if ( !strcmp(c->key, "pwfile") ) {
    pwmode = PWMODE_DSTR;
    password = c->value;
   } else if ( !strcmp(c->key, "codec") ) {
    codec = c->value;
   } else {
    ROAR_WARN("Unknown output option: %s=%s", c->key, c->value);
   }
  }

  roar_mm_free(kv);
 }

 switch (pwmode) {
  case PWMODE_NONE:
    password = NULL;
   break;
  case PWMODE_ARG:
    // Noop
   break;
  case PWMODE_ASK:
    if ( roar_passwd_simple_ask_pw(&pwbuf, "Password for icecast server?", NULL) == -1 ) {
     return -1;
    }
    password = pwbuf;
   break;
  case PWMODE_DSTR:
    vio = roar_vio_open_dstr_simple_new(password, O_RDONLY);
    if ( vio == NULL )
     return -1;

    password = read_pw(vio);

    roar_vio_close(vio);
   break;
 }

 mixer = savannahice_mixer_new_simple(server, port, mount, user, password);
 if ( pwbuf != NULL )
  roar_mm_free(pwbuf);
 if ( mixer == NULL )
  return -1;

 if ( codec != NULL ) {
  if ( savannahice_mixer_set_codec(mixer, roar_str2codec(codec)) == -1 ) {
   savannahice_mixer_unref(mixer);
   return -1;
  }
 }

 ret = savannahice_mixerstore_add_mixer(g_mixerstore, mixer);
 savannahice_mixer_unref(mixer);

 return ret;
}

#define _CKHAVEARGS(x) if ( (i + (x)) >= argc ) { \
                        ROAR_ERR("Option %s requires more arguments. See --help for more details.", k); \
                        savannahice_exit(70); \
                       }

int main (int argc, char * argv[]) {
 char * m_opts = NULL;
 const char * m_dev = NULL;
 const char * server_address = NULL; // DEFAULT_SERVER_ADDRESS is handled within the network object.
 int          server_type    = DEFAULT_SERVER_TYPE;
 int          server_port    = 0;
 int          server_proto   = ROAR_PROTO_ROARAUDIO;
 int no_listen = 0;
 const char * plugin_args = NULL;
 const char * k;
 int i;
 int_least32_t sleep_time;

#ifdef SIGPIPE
 signal(SIGPIPE, SIG_IGN);  // ignore broken pipes
#endif

 shout_init();

 if ( savannahice_stream_init() != 0 ) {
  ROAR_ERR("Can not init internal stream structures. Very bad.");
  savannahice_exit(71);
 }

 g_plugins = savannahice_plugin_new();
 g_network = savannahice_network_new();
 g_mixerstore = savannahice_mixerstore_new();

// savannahice_network_add_protocol(g_network, &g_proto_roar);
 roar_dl_register_fn(ROAR_DL_HANDLE_APPLICATION, ROAR_DL_FN_PROTO, ROAR_DL_PROTO_SUBTYPE,
                     &g_proto_roar_impl, sizeof(g_proto_roar_impl), ROAR_DL_PROTO_VERSION, ROAR_DL_FNREG_OPT_NONE);

 if ( g_plugins == NULL || g_network == NULL || g_mixerstore == NULL ) {
  ROAR_ERR("Can not init internal structures. Very bad.");
  savannahice_exit(71);
 }

 for (i = 1; i < argc; i++) {
  k = argv[i];

  if ( !strcmp(k, "--help") ) {
   usage(argv[0]);
   savannahice_exit(0);
  } else if ( strcmp(k, "--verbose") == 0 ) {
   g_verbose++;

  } else if ( strcmp(k, "--plugin-load") == 0 ) {
   _CKHAVEARGS(1);
   if ( savannahice_plugin_load(g_plugins, argv[++i], plugin_args) == -1 ) {
    ROAR_ERR("Can not load plugin");
   }
   plugin_args = NULL;
  } else if ( strcmp(k, "--plugin-args") == 0 ) {
   _CKHAVEARGS(1);
   plugin_args = argv[++i];

/*
X -o  --odriver DRV     - Set the driver, use '--list-driver' to get a list
X -O  --odevice DEV     - Set the device
X -oO OPTS              - Set output options
X -oN                   - Adds another output
 -oP                   - Mark output as primary
 --list-driver         - List all drivers (default driver: (autodetect))
*/
  } else if ( strcmp(k, "-o") == 0 || strcmp(k, "--odriver") == 0 ) {
   _CKHAVEARGS(1);
   i++;
   if ( !!strcmp(argv[i], "shout") ) {
    ROAR_ERR("Unknown output driver: %s", argv[i]);
    savannahice_exit(1);
   }
  } else if ( strcmp(k, "-O") == 0 || strcmp(k, "--odevice") == 0 ) {
   _CKHAVEARGS(1);
   m_dev = argv[++i];
  } else if ( strcmp(k, "-oO") == 0 ) {
   _CKHAVEARGS(1);
   m_opts = argv[++i];
  } else if ( strcmp(k, "-oN") == 0 ) {
   if ( savannahice_add_mixer(m_opts, m_dev) == -1 ) {
    fprintf(stderr, "Error: can not open stream: %s: %s\n",
                    m_dev == NULL ? "(default server)" : m_dev, roar_error2str(roar_error));
    savannahice_exit(1);
   }
   m_opts = NULL;
   m_dev  = NULL;
  } else if ( strcmp(k, "--list-driver") == 0 ) {
   print_driverlist();
   savannahice_exit(0);

  } else if ( strcmp(k, "--bind") == 0 ) {
   _CKHAVEARGS(1);
   server_address = argv[++i];
  } else if ( strcmp(k, "--port") == 0 ) {
   _CKHAVEARGS(1);
   server_port = atoi(argv[++i]);
  } else if ( strcmp(k, "--tcp") == 0 || strcmp(k, "-t") == 0 ) {
   server_type = ROAR_SOCKET_TYPE_TCP;
  } else if ( strcmp(k, "--unix") == 0 || strcmp(k, "-u") == 0 ) {
   server_type = ROAR_SOCKET_TYPE_UNIX;
   server_port = 0;
  } else if ( strcmp(k, "--decnet") == 0 || strcmp(k, "-n") == 0 ) {
   server_type = ROAR_SOCKET_TYPE_DECNET;
   server_port = 0;
  } else if ( strcmp(k, "--proto") == 0 ) {
   if ( (server_proto = roar_str2proto(argv[++i])) == -1 ) {
    ROAR_ERR("Unknown protocol: %s", argv[i]);
    return 1;
   }
  } else if ( strcmp(k, "--new-sock") == 0 ) {
   savannahice_network_add_listen_address(g_network, server_type, server_address, server_port, server_proto);
  } else if ( strcmp(k, "--no-listen") == 0 ) {
   no_listen = 1;
  } else if ( strcmp(k, "--client-fh") == 0 ) {
   _CKHAVEARGS(1);
   if ( savannahice_network_add_client_fh(g_network, atoi(argv[++i]), ROAR_PROTO_ROARAUDIO) == -1 ) {
    ROAR_ERR("Can not add client given via --client-fh: %s", roar_error2str(roar_error));
   }

  } else if ( strcmp(k, "--close-fh") == 0 ) {
   _CKHAVEARGS(1);
   close(atoi(argv[++i]));

  } else {
   fprintf(stderr, "Error: Invalid parameter: %s\n", k);
   usage(argv[0]);
   savannahice_exit(1);
   return 1;
  }
 }

 if ( savannahice_add_mixer(m_opts, m_dev) == -1 ) {
  fprintf(stderr, "Error: can not open stream: %s: %s\n",
                  m_dev == NULL ? "(default server)" : m_dev, roar_error2str(roar_error));
  return 1;
 }

 if ( !no_listen ) {
  savannahice_network_add_listen_address(g_network, server_type, server_address, server_port, server_proto);
 }

 sleep_time = DEFAULT_SLEEP_TIME;
 while (savannahice_network_check(g_network, sleep_time) != -1) {
  savannahice_mixerstore_handle(g_mixerstore);

  sleep_time = savannahice_mixerstore_delay(g_mixerstore);
  if ( sleep_time < 1 )
   sleep_time = DEFAULT_SLEEP_TIME;

  savannahice_plugin_trigger(g_plugins, ROAR_DL_APPSCHED_UPDATE);
 }

 savannahice_exit(0);
 return 0;
}

//ll
