//stream.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"

#define MAX_BUFFER_PARTS 16
#define MAX_BUFFER_SIZE  (1024*8)

savannahice_pool_t __pool = NULL;

int savannahice_stream_init(void) {
 __pool = savannahice_pool_new(-1);
 if ( __pool == NULL )
  return -1;

 return 0;
}

int savannahice_stream_free(void) {
 return savannahice_pool_unref(&__pool);
}

// This function MUST NOT set error codes!
static inline void __free_id(int id) {
 int err = roar_error;
 savannahice_pool_id_free(&__pool, id);
 roar_err_set(err);
}
//static int __isused_id(int id);

struct savannahice_stream {
 size_t refc;
 int id;
 int client;
 int dir;
 uint32_t codec;
 struct roar_buffer * buffer;
 int status; // ROAR_STREAMSTATE_*
};

savannahice_stream_t savannahice_stream_new(int client, int dir, uint32_t codec) {
 savannahice_stream_t ret;
 int id = savannahice_pool_id_new(&__pool);

 if ( id == -1 )
  return NULL;

 ret = roar_mm_malloc(sizeof(struct savannahice_stream));

 if ( ret == NULL ) {
  __free_id(id);
  return NULL;
 }

 memset(ret, 0, sizeof(*ret));

 ret->refc = 1;
 ret->id = id;
 ret->client = client;
 ret->dir = dir;
 ret->codec = codec;
 ret->status = ROAR_STREAMSTATE_NEW;
 ret->buffer = NULL;

 return ret;
}

int savannahice_stream_ref(savannahice_stream_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc++;

 return 0;
}

int savannahice_stream_unref(savannahice_stream_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc--;

 if ( self->refc == 0 ) {
  __free_id(self->id);

  if ( self->buffer != NULL )
   roar_buffer_free(self->buffer);
  roar_mm_free(self);
 }

 return 0;
}

#define _getter(type,member) \
type savannahice_stream_get_ ## member (savannahice_stream_t self) { \
 if ( self == NULL ) { \
  roar_err_set(ROAR_ERROR_FAULT); \
  return -1; \
 } \
 \
 return self->member; \
}

_getter(int,id)
_getter(int,client)
_getter(int,dir)
_getter(int,status)
_getter(uint32_t,codec)

int savannahice_stream_set_status(savannahice_stream_t self, const int status) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->status = status;

 return 0;
}

int savannahice_stream_add_buffer(savannahice_stream_t self, struct roar_buffer * buffer) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( self->buffer == NULL ) {
  if ( roar_buffer_ref(buffer) == -1 )
   return -1;
  self->buffer = buffer;
  return 0;
 } else {
  return roar_buffer_add(self->buffer, buffer);
 }
}

struct roar_buffer ** savannahice_stream_get_buffer(savannahice_stream_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 return &(self->buffer);
}

ssize_t savannahice_stream_get_bufferspace(savannahice_stream_t self) {
 struct roar_buffer_stats stats;

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 if ( self->buffer == NULL )
  return MAX_BUFFER_SIZE;

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 if ( roar_buffer_ring_stats(self->buffer, &stats) == -1 )
  return -1;

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 if ( stats.parts >= MAX_BUFFER_PARTS )
  return 0;

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 if ( stats.bytes >= MAX_BUFFER_SIZE )
  return 0;

 ROAR_DBG("savannahice_stream_get_bufferspace(self=%p) = ?", self);

 return MAX_BUFFER_SIZE - stats.bytes;
}

int savannahice_stream_get_alone(savannahice_stream_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( self->refc > 1 )
  return 0;

 return 1;
}

//ll
